from .median import CoordinateWiseMedian
from .trimmed_mean import CoordinateWiseTrimmedMean
from .mean_of_medians import MeanOfMedians
from . import _tiling  # ensure flatten helpers import when package is loaded
